import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateWorkspaceEverythingLevelViewRequest(BaseModel):
    """Request schema for `CreateWorkspaceEverythingLevelView`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    name: str = Field(
        default=...,
        alias="name",
        description="Name",
    )
    type: str = Field(
        default=...,
        alias="type",
        description=(
            "The type of view to create. Options include: `list`, `board`, `calendar`, "
            "`table`, `timeline`, `workload`, `activity`, `map`, `conversation`, or `gantt`. "
        ),
    )
    grouping_field: str = Field(
        default=...,
        alias="grouping__field",
        description=(
            "Set the field to group by.   Options include: `none`, `status`, `priority`, "
            "`assignee`, `tag`, or `dueDate`. "
        ),
    )
    grouping_dir: int = Field(
        default=...,
        alias="grouping__dir",
        description=(
            "Set a group sort order using `1` or `-1`.   For example, use `1`show tasks "
            "with urgent priority at the top of your view, and tasks with no priority "
            "at the bottom.   Use `-1` to reverse the order to show tasks with no priority "
            "at the top of your view. "
        ),
    )
    grouping_collapsed: t.List[str] = Field(
        default=...,
        alias="grouping__collapsed",
        description="",
    )
    grouping_ignore: bool = Field(
        default=...,
        alias="grouping__ignore",
        description="Ignore",
    )
    divide_field: None = Field(
        default=None,
        alias="divide__field",
        description="Field",
    )
    divide_dir: None = Field(
        default=None,
        alias="divide__dir",
        description="Dir",
    )
    divide_collapsed: bool = Field(
        default=...,
        alias="divide__collapsed",
        description="Collapsed",
    )
    sorting_fields: t.List[str] = Field(
        default=...,
        alias="sorting__fields",
        description=(
            "Include an array of fields to sort by.    You can sort by the same fields "
            "available when [filtering a view](https://clickup.com/api). "
        ),
    )
    filters_op: str = Field(
        default=...,
        alias="filters__op",
        description="The available operator (`op``) values are `AND`` and `OR``.",
    )
    filters_fields: t.List[str] = Field(
        default=...,
        alias="filters__fields",
        description=(
            "View the list of [fields available](https://clickup.com/api) to filter by. "
        ),
    )
    filters_search: str = Field(
        default=...,
        alias="filters__search",
        description="Search",
    )
    filters_show_closed: bool = Field(
        default=...,
        alias="filters__show_closed",
        description="Show Closed",
    )
    columns_fields: t.List[str] = Field(
        default=...,
        alias="columns__fields",
        description=(
            "Custom Fields require the `_cf` prefix and must be formatted as a JSON object. "
        ),
    )
    team_sidebar_assignees: t.List[str] = Field(
        default=...,
        alias="team_sidebar__assignees",
        description="",
    )
    team_sidebar_assigned_comments: bool = Field(
        default=...,
        alias="team_sidebar__assigned_comments",
        description="Assigned Comments",
    )
    team_sidebar_unassigned_tasks: bool = Field(
        default=...,
        alias="team_sidebar__unassigned_tasks",
        description="Unassigned Tasks",
    )
    settings_show_task_locations: bool = Field(
        default=...,
        alias="settings__show_task_locations",
        description="Show Task Locations",
    )
    settings_show_subtasks: int = Field(
        default=...,
        alias="settings__show_subtasks",
        description=(
            "Acceptable values are `1`, `2`, or `3`, which show subtasks separate, expanded, "
            "or collapsed. "
        ),
    )
    settings_show_subtask_parent_names: bool = Field(
        default=...,
        alias="settings__show_subtask_parent_names",
        description="Show Subtask Parent Names",
    )
    settings_show_closed_subtasks: bool = Field(
        default=...,
        alias="settings__show_closed_subtasks",
        description="Show Closed Subtasks",
    )
    settings_show_assignees: bool = Field(
        default=...,
        alias="settings__show_assignees",
        description="Show Assignees",
    )
    settings_show_images: bool = Field(
        default=...,
        alias="settings__show_images",
        description="Show Images",
    )
    settings_collapse_empty_columns: str = Field(
        default=...,
        alias="settings__collapse_empty_columns",
        description="Collapse Empty Columns",
    )
    settings_me_comments: bool = Field(
        default=...,
        alias="settings__me_comments",
        description="Me Comments",
    )
    settings_me_subtasks: bool = Field(
        default=...,
        alias="settings__me_subtasks",
        description="Me Subtasks",
    )
    settings_me_checklists: bool = Field(
        default=...,
        alias="settings__me_checklists",
        description="Me Checklists",
    )


class CreateWorkspaceEverythingLevelViewResponse(BaseModel):
    """Response schema for `CreateWorkspaceEverythingLevelView`"""

    data: t.Dict[str, t.Any]


class CreateWorkspaceEverythingLevelView(OpenAPIAction):
    """
    Add a List, Board, Calendar, Table, Timeline, Workload, Activity, Map, Chat,
    or Gantt view at the Everything Level of a Workspace.
    """

    _tags = ["Views"]
    _display_name = "create_workspace_everything_level_view"
    _request_schema = CreateWorkspaceEverythingLevelViewRequest
    _response_schema = CreateWorkspaceEverythingLevelViewResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/view"
    method = "post"
    operation_id = "Views_createWorkspaceViewEverythingLevel"
    action_identifier = "/team/{team_id}/view_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "name": {"__alias": "name"},
        "type": {"__alias": "type"},
        "grouping": {
            "__alias": "grouping",
            "field": {"__alias": "field"},
            "dir": {"__alias": "dir"},
            "collapsed": {"__alias": "collapsed"},
            "ignore": {"__alias": "ignore"},
        },
        "divide": {
            "__alias": "divide",
            "field": {"__alias": "field"},
            "dir": {"__alias": "dir"},
            "collapsed": {"__alias": "collapsed"},
        },
        "sorting": {"__alias": "sorting", "fields": {"__alias": "fields"}},
        "filters": {
            "__alias": "filters",
            "op": {"__alias": "op"},
            "fields": {"__alias": "fields"},
            "search": {"__alias": "search"},
            "show_closed": {"__alias": "show_closed"},
        },
        "columns": {"__alias": "columns", "fields": {"__alias": "fields"}},
        "team_sidebar": {
            "__alias": "team_sidebar",
            "assignees": {"__alias": "assignees"},
            "assigned_comments": {"__alias": "assigned_comments"},
            "unassigned_tasks": {"__alias": "unassigned_tasks"},
        },
        "settings": {
            "__alias": "settings",
            "show_task_locations": {"__alias": "show_task_locations"},
            "show_subtasks": {"__alias": "show_subtasks"},
            "show_subtask_parent_names": {"__alias": "show_subtask_parent_names"},
            "show_closed_subtasks": {"__alias": "show_closed_subtasks"},
            "show_assignees": {"__alias": "show_assignees"},
            "show_images": {"__alias": "show_images"},
            "collapse_empty_columns": {"__alias": "collapse_empty_columns"},
            "me_comments": {"__alias": "me_comments"},
            "me_subtasks": {"__alias": "me_subtasks"},
            "me_checklists": {"__alias": "me_checklists"},
        },
    }

    aliases = {
        "grouping": "fb8a835e",
        "divide": "d01a3e45",
        "sorting": "24c3a5b6",
        "filters": "c2edf54f",
        "columns": "d19274c3",
        "team_sidebar": "3a20b9ce",
        "settings": "800f8573",
    }



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class AddTagsFromTimeEntriesRequest(BaseModel):
    """Request schema for `AddTagsFromTimeEntries`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    tags: t.List[dict] = Field(
        default=...,
        alias="tags",
        description="",
    )
    time_entry_ids: t.List[str] = Field(
        default=...,
        alias="time_entry_ids",
        description="",
    )


class AddTagsFromTimeEntriesResponse(BaseModel):
    """Response schema for `AddTagsFromTimeEntries`"""

    data: t.Dict[str, t.Any]


class AddTagsFromTimeEntries(OpenAPIAction):
    """Add a label to a time entry."""

    _tags = ["Time Tracking"]
    _display_name = "add_tags_from_time_entries"
    _request_schema = AddTagsFromTimeEntriesRequest
    _response_schema = AddTagsFromTimeEntriesResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/tags"
    method = "post"
    operation_id = "TimeTracking_addTagsFromTimeEntries"
    action_identifier = "/team/{team_id}/time_entries/tags_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "tags": {"__alias": "tags"},
        "time_entry_ids": {"__alias": "time_entry_ids"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetRunningTimeEntryRequest(BaseModel):
    """Request schema for `GetRunningTimeEntry`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    assignee: int = Field(
        default=None,
        alias="assignee",
        description="user id",
    )


class GetRunningTimeEntryResponse(BaseModel):
    """Response schema for `GetRunningTimeEntry`"""

    data: t.Dict[str, t.Any]


class GetRunningTimeEntry(OpenAPIAction):
    """
    View a time entry that's currently tracking time for the authenticated user.
    ***Note:** A time entry that has a negative duration means that timer
    is currently running for that user.*
    """

    _tags = ["Time Tracking"]
    _display_name = "get_running_time_entry"
    _request_schema = GetRunningTimeEntryRequest
    _response_schema = GetRunningTimeEntryResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/current"
    method = "get"
    operation_id = "TimeTracking_getCurrentTimeEntry"
    action_identifier = "/team/{team_id}/time_entries/current_get"

    path_params = {"team_id": "team_id"}
    query_params = {"assignee": "assignee"}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class RemoveGuestFromWorkspaceRequest(BaseModel):
    """Request schema for `RemoveGuestFromWorkspace`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    guest_id: int = Field(
        ...,
        alias="guest_id",
        description="",
    )


class RemoveGuestFromWorkspaceResponse(BaseModel):
    """Response schema for `RemoveGuestFromWorkspace`"""

    data: t.Dict[str, t.Any]


class RemoveGuestFromWorkspace(OpenAPIAction):
    """
    Revoke a guest's access to a Workspace.    ***Note:** This endpoint is only
    available to Workspaces on our [Enterprise Plan](https://clickup.com/pricing).*
    """

    _tags = ["Guests"]
    _display_name = "remove_guest_from_workspace"
    _request_schema = RemoveGuestFromWorkspaceRequest
    _response_schema = RemoveGuestFromWorkspaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/guest/{guest_id}"
    method = "delete"
    operation_id = "Guests_revokeGuestAccessToWorkspace"
    action_identifier = "/team/{team_id}/guest/{guest_id}_delete"

    path_params = {"team_id": "team_id", "guest_id": "guest_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
Integrations manager for Composio SDK.

Usage:
    composio connections [command] [options]
"""

import click

from composio.cli.context import Context, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase
from composio.core.cls.did_you_mean import DYMGroup


class ConnectionsExamples(HelpfulCmdBase, DYMGroup):
    examples = [
        click.style("composio connections", fg="green")
        + click.style("             # List all connections\n", fg="black"),
        click.style("composio connections get 123", fg="green")
        + click.style("     # Get details of connection with ID 123\n", fg="black"),
        click.style("composio connections delete 456", fg="green")
        + click.style("  # Delete connection with ID 456\n", fg="black"),
    ]


@click.group(name="connections", invoke_without_command=True, cls=ConnectionsExamples)
@click.option(
    "--active",
    help="Show only active connections",
    is_flag=True,
)
@click.help_option("--help", "-h", "-help")
@pass_context
def _connections(context: Context, active: bool = False) -> None:
    """List composio connections for your account"""
    if context.click_ctx.invoked_subcommand:
        return

    for connection in context.client.connected_accounts.get(active=active):
        context.console.print(f"• Id : {connection.id}")
        context.console.print(f"  App: {connection.appUniqueId}")


class GetExamples(HelpfulCmdBase, click.Command):
    examples = [
        click.style("composio connections get 123", fg="green")
        + click.style("  # Get details of connection with ID 123\n", fg="black"),
    ]


@_connections.command(name="get", cls=GetExamples)
@click.argument("id", type=str)
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _get(context: Context, id: str) -> None:
    """Get connection information"""
    connection = context.client.get_entity().get_connection(
        connected_account_id=id,
    )
    context.console.print(f"[green]App   :[/green] {connection.appUniqueId}")
    context.console.print(f"[green]Id    :[/green] {connection.id}")
    context.console.print(f"[green]Status:[/green] {connection.status}")



import json
import typing as t

import click

from composio.cli.context import Context, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.client import Action


@click.command(name="execute")
@click.argument("action", type=str)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Action parameters as a JSON string",
)
@click.option(
    "-m",
    "--metadata",
    type=str,
    help="Metadata for executing an action",
)
@handle_exceptions(
    {"cls": json.JSONDecodeError, "message": "Invalid JSON format for parameters"}
)
@pass_context
def _execute(
    context: Context,
    action: str,
    params: t.Optional[str] = None,
    metadata: t.Optional[str] = None,
) -> None:
    """Execute a Composio action"""
    context.console.print(
        json.dumps(
            obj=context.toolset.execute_action(
                action=Action(action),
                params=json.loads(params) if params else {},
                metadata=json.loads(metadata) if metadata else {},
            )
        )
    )



"""
Login utility for composio SDK.

Usage:
    composio login
"""

import click

from composio.cli.context import Context, login, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase


class Examples(HelpfulCmdBase, click.Command):
    examples = [
        click.style("composio login --help", fg="green")
        + click.style("  # Display help for login command\n", fg="black"),
        click.style("composio login --no-browser", fg="green")
        + click.style("  # Login without browser interaction\n", fg="black"),
    ]


@click.command(name="login", cls=Examples)
@click.option(
    "--no-browser",
    is_flag=True,
    default=False,
    help="Prevent from opening browser window",
)
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _login(
    context: Context,
    no_browser: bool = False,
) -> None:
    """Login to the Composio SDK"""

    if context.is_logged_in():
        context.console.print("\n[green]✔ You're already logged in![/green]\n")
        context.console.print(
            "> Use [green]'composio logout'[/green] to log out and then login again"
        )
        return

    if context.using_api_key_from_env():
        context.console.print(
            "[yellow]WARNING: `COMPOSIO_API_KEY` Environment variable is set[/yellow]"
        )

    context.console.print("\n> [green]Authenticating...[/green]")
    login(context=context, no_browser=no_browser)



"""
Composio SDK client.
"""

import os
import sys
import typing as t
from datetime import datetime
from pathlib import Path

import requests

from composio.client.collections import (
    AUTH_SCHEMES,
    Actions,
    ActiveTriggerModel,
    ActiveTriggers,
    Apps,
    AuthSchemeType,
    ConnectedAccountModel,
    ConnectedAccounts,
    ConnectionRequestModel,
    CustomAuthObject,
    IntegrationModel,
    Integrations,
    Logs,
    Triggers,
)
from composio.client.endpoints import v1
from composio.client.enums import (
    Action,
    App,
    AppType,
    Tag,
    TagType,
    Trigger,
    TriggerType,
)
from composio.client.exceptions import ComposioClientError, HTTPError, NoItemsFound
from composio.client.http import HttpClient
from composio.constants import (
    DEFAULT_ENTITY_ID,
    ENV_COMPOSIO_API_KEY,
    LOCAL_CACHE_DIRECTORY_NAME,
    USER_DATA_FILE_NAME,
)
from composio.exceptions import ApiKeyNotProvidedError
from composio.storage.user import UserData
from composio.utils.url import get_api_url_base


_valid_keys: t.Set[str] = set()
_clients: t.List["Composio"] = []


class Composio:
    """Composio SDK Client."""

    local: t.Any
    _api_key: t.Optional[str] = None
    _http: t.Optional[HttpClient] = None

    def __init__(
        self,
        api_key: t.Optional[str] = None,
        base_url: t.Optional[str] = None,
        runtime: t.Optional[str] = None,
    ) -> None:
        """
        Initialize Composio SDK client

        :param api_key: Authentication key for Composio server
        :param base_url: Base URL for Composio server
        :param runtime: Runtime specifier
        """
        self._api_key = api_key
        self.runtime = runtime
        self.base_url = base_url or get_api_url_base()

        self.apps = Apps(client=self)
        self.actions = Actions(client=self)
        self.triggers = Triggers(client=self)
        self.integrations = Integrations(client=self)
        self.active_triggers = ActiveTriggers(client=self)
        self.connected_accounts = ConnectedAccounts(client=self)
        self.logs = Logs(client=self)
        _clients.append(self)

    @staticmethod
    def get_latest() -> "Composio":
        """Get latest composio client from the runtime stack."""
        if len(_clients) == 0:
            _ = Composio()
        return _clients[-1]

    @property
    def api_key(self) -> str:
        if self._api_key is None:
            cache_dir = Path.home() / LOCAL_CACHE_DIRECTORY_NAME
            user_data_path = cache_dir / USER_DATA_FILE_NAME
            user_data = (
                UserData.load(path=user_data_path) if user_data_path.exists() else None
            )
            env_api_key = (
                user_data.api_key if user_data else os.environ.get(ENV_COMPOSIO_API_KEY)
            )
            if env_api_key:
                self._api_key = env_api_key

        if self._api_key is None:
            raise ApiKeyNotProvidedError()

        self._api_key = self.validate_api_key(
            key=t.cast(str, self._api_key),
            base_url=self.base_url,
        )

        return self._api_key

    @api_key.setter
    def api_key(self, value: str) -> None:
        self._api_key = value

    @property
    def http(self) -> HttpClient:
        if not self._http:
            self._http = HttpClient(
                base_url=self.base_url,
                api_key=self.api_key,
                runtime=self.runtime,
            )
        return self._http

    @http.setter
    def http(self, value: HttpClient) -> None:
        self._http = value

    @staticmethod
    def validate_api_key(key: str, base_url: t.Optional[str] = None) -> str:
        """Validate given API key."""
        if key in _valid_keys:
            return key

        base_url = base_url or get_api_url_base()
        response = requests.get(
            url=base_url + str(v1 / "client" / "auth" / "client_info"),
            headers={
                "x-api-key": key,
            },
            timeout=60,
        )
        if response.status_code in (401, 403):
            raise ComposioClientError("API Key is not valid!")

        if response.status_code != 200:
            raise ComposioClientError(f"Unexpected error: HTTP {response.status_code}")

        _valid_keys.add(key)
        return key

    @staticmethod
    def generate_auth_key(base_url: t.Optional[str] = None) -> str:
        """Generate auth key."""
        http = HttpClient(
            base_url=base_url or get_api_url_base(),
            api_key="",
        )
        response = http.get(url=str(v1.cli.generate_cli_session))
        if response.status_code != 200:
            raise HTTPError(
                message=response.content.decode(),
                status_code=response.status_code,
            )
        data = response.json()
        return data["key"]

    @staticmethod
    def validate_auth_session(
        key: str,
        code: str,
        base_url: t.Optional[str] = None,
    ) -> str:
        """
        Validate API session.

        :param key: Session key
        :param code: Authentication code
        """
        http = HttpClient(
            base_url=base_url or get_api_url_base(),
            api_key="",
        )
        response = http.get(str(v1.cli.verify_cli_code({"key": key, "code": code})))
        if response.status_code != 200:
            raise HTTPError(
                message=response.content.decode(),
                status_code=response.status_code,
            )
        data = response.json()
        return data["apiKey"]

    def get_entity(self, id: str = DEFAULT_ENTITY_ID) -> "Entity":
        """
        Create Entity object.

        :param id: Entity ID
        :return: Entity object.
        """
        return Entity(id=id, client=self)


class Entity:
    """Class to represent Entity object."""

    def __init__(
        self,
        client: Composio,
        id: str = DEFAULT_ENTITY_ID,
    ) -> None:
        """
        Initialize Entity object.

        :param client: Composio client object.
        :param id: Entity ID string
        """
        self.client = client
        self.id = id

    def execute(
        self,
        action: Action,
        params: t.Dict,
        connected_account_id: t.Optional[str] = None,
        session_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
        auth: t.Optional[CustomAuthObject] = None,
    ) -> t.Dict:
        """
        Execute an action.

        :param action: Action ID (Enum)
        :param params: Parameters for executing actions
        :param connected_account_id: Connection ID if you want to use a specific
                connection
        :param session_id: ID of the current workspace session
        :return: Dictionary containing execution result
        """
        if action.no_auth:
            return self.client.actions.execute(
                action=action,
                params=params,
                entity_id=self.id,
                session_id=session_id,
                text=text,
            )

        if auth is not None:
            return self.client.actions.execute(
                action=action,
                params=params,
                entity_id=self.id,
                session_id=session_id,
                text=text,
                auth=auth,
            )

        connected_account = self.get_connection(
            app=action.app,
            connected_account_id=connected_account_id,
        )

        return self.client.actions.execute(
            action=action,
            params=params,
            entity_id=t.cast(str, connected_account.clientUniqueUserId),
            connected_account=connected_account.id,
            session_id=session_id,
            text=text,
            auth=auth,
        )

    def get_connection(
        self,
        app: t.Optional[AppType] = None,
        connected_account_id: t.Optional[str] = None,
    ) -> ConnectedAccountModel:
        """
        Get connected account for an action.

        :param app: App name
        :param connected_account_id: Connected account ID to use as filter
        :return: Connected account object
        :raises: If no connected account found for given entity ID
        """
        if connected_account_id is not None:
            return self.client.connected_accounts.get(
                connection_id=connected_account_id
            )

        latest_account = None
        latest_creation_date = datetime.fromtimestamp(0.0)
        connected_accounts = self.client.connected_accounts.get(
            entity_ids=[self.id],
            active=True,
        )
        app = str(app).lower()
        for connected_account in connected_accounts:
            if app == connected_account.appUniqueId:
                creation_date = datetime.fromisoformat(
                    connected_account.createdAt.replace("Z", "+00:00")
                )
                if latest_account is None or creation_date > latest_creation_date:
                    latest_creation_date = creation_date
                    latest_account = connected_account

        if latest_account is None:
            entity = self.id
            suggestion = (
                f"composio add {app}"
                if entity == DEFAULT_ENTITY_ID
                else f"composio add {app} -e {entity}"
            )
            note = f"Run this command to create a new connection: {suggestion}"
            doc_note = "Read more here: https://dub.composio.dev/auth-help"
            if sys.version_info >= (3, 11):
                exception = NoItemsFound(
                    f"Could not find a connection with {app=},"
                    f" {connected_account_id=} and {entity=}."
                )
                exception.add_note(note)
                exception.add_note(doc_note)
            else:
                exception = NoItemsFound(
                    f"Could not find a connection with {app=},"
                    f" {connected_account_id=} and {entity=}.\n{note}\n{doc_note}"
                )
            raise exception

        return latest_account

    def get_connections(self) -> t.List[ConnectedAccountModel]:
        """
        Get all connections for an entity.
        """
        return self.client.connected_accounts.get(entity_ids=[self.id], active=True)

    def enable_trigger(
        self, app: t.Union[str, App], trigger_name: str, config: t.Dict[str, t.Any]
    ) -> t.Dict:
        """
        Enable a trigger for an entity.

        :param app: App name
        :param trigger_name: Trigger name
        :param config: Trigger config
        """
        connected_account = self.get_connection(app=app)
        return self.client.triggers.enable(
            name=trigger_name,
            connected_account_id=connected_account.id,
            config=config,
        )

    def disable_trigger(self, trigger_id: str) -> t.Dict:
        """
        Disable a trigger for an entity.

        :param trigger_id: Trigger ID
        """
        return self.client.triggers.disable(id=trigger_id)

    def get_active_triggers(self) -> t.List[ActiveTriggerModel]:
        """
        Get all active triggers for an entity.
        """
        connected_accounts = self.get_connections()
        return self.client.active_triggers.get(
            connected_account_ids=[
                connected_account.id for connected_account in connected_accounts
            ]
        )

    def initiate_connection(
        self,
        app_name: t.Union[str, App],
        auth_mode: t.Optional[str] = None,
        auth_config: t.Optional[t.Dict[str, t.Any]] = None,
        redirect_url: t.Optional[str] = None,
        integration: t.Optional[IntegrationModel] = None,
        use_composio_auth: bool = True,
        force_new_integration: bool = False,
        connected_account_params: t.Optional[t.Dict] = None,
        labels: t.Optional[t.List] = None,
    ) -> ConnectionRequestModel:
        """
        Initiate an integration connection process for a specified application.

        :param app_name: The name of the application or an App enum instance.
        :param auth_mode: Optional authentication mode to be used.
        :param auth_config: Optional dictionary containing authentication configuration details.
        :param redirect_url: Optional URL to which a user will be redirected after authentication.
        :param integration: Optional existing IntegrationModel instance to be used.
        :return: A ConnectionRequestModel instance representing the initiated connection.
        """
        if isinstance(app_name, App):
            app_name = app_name.slug

        app = self.client.apps.get(name=app_name)
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        if integration is None and auth_mode is not None:
            if auth_mode not in AUTH_SCHEMES:
                raise ComposioClientError(
                    f"'auth_mode' should be one of {AUTH_SCHEMES}"
                )
            auth_mode = t.cast(AuthSchemeType, auth_mode)
            if "OAUTH" not in auth_mode:
                use_composio_auth = False
            integration = self.client.integrations.create(
                app_id=app.appId,
                name=f"{app_name}_{timestamp}",
                auth_mode=auth_mode,
                auth_config=auth_config,
                use_composio_auth=use_composio_auth,
                force_new_integration=force_new_integration,
            )

        if integration is None and auth_mode is None:
            integration = self.client.integrations.create(
                app_id=app.appId,
                auth_config=auth_config,
                name=f"{app_name}_{timestamp}",
                use_composio_auth=use_composio_auth,
                force_new_integration=force_new_integration,
            )

        return self.client.connected_accounts.initiate(
            integration_id=t.cast(IntegrationModel, integration).id,
            entity_id=self.id,
            params=connected_account_params,
            labels=labels,
            redirect_url=redirect_url,
        )


__all__ = (
    "Action",
    "App",
    "Tag",
    "AppType",
    "TagType",
    "Trigger",
    "TriggerType",
    "Composio",
)



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateChatViewCommentRequest(BaseModel):
    """Request schema for `CreateChatViewComment`"""

    view_id: str = Field(
        ...,
        alias="view_id",
        description="105 (string)",
    )
    comment_text: str = Field(
        default=...,
        alias="comment_text",
        description="Comment Text",
    )
    notify_all: bool = Field(
        default=...,
        alias="notify_all",
        description=(
            "If `notify_all` is true, notifications will be sent to everyone including "
            "the creator of the comment. "
        ),
    )


class CreateChatViewCommentResponse(BaseModel):
    """Response schema for `CreateChatViewComment`"""

    data: t.Dict[str, t.Any]


class CreateChatViewComment(OpenAPIAction):
    """Add a new comment to a Chat view."""

    _tags = ["Comments"]
    _display_name = "create_chat_view_comment"
    _request_schema = CreateChatViewCommentRequest
    _response_schema = CreateChatViewCommentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/view/{view_id}/comment"
    method = "post"
    operation_id = "Comments_createChatViewComment"
    action_identifier = "/view/{view_id}/comment_post"

    path_params = {"view_id": "view_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "comment_text": {"__alias": "comment_text"},
        "notify_all": {"__alias": "notify_all"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateTaskCommentRequest(BaseModel):
    """Request schema for `CreateTaskComment`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )
    comment_text: str = Field(
        default=...,
        alias="comment_text",
        description="Comment Text",
    )
    assignee: int = Field(
        default=...,
        alias="assignee",
        description="Assignee",
    )
    notify_all: bool = Field(
        default=...,
        alias="notify_all",
        description=(
            "If `notify_all` is true, notifications will be sent to everyone including "
            "the creator of the comment. "
        ),
    )


class CreateTaskCommentResponse(BaseModel):
    """Response schema for `CreateTaskComment`"""

    data: t.Dict[str, t.Any]


class CreateTaskComment(OpenAPIAction):
    """Add a new comment to a task."""

    _tags = ["Comments"]
    _display_name = "create_task_comment"
    _request_schema = CreateTaskCommentRequest
    _response_schema = CreateTaskCommentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/comment"
    method = "post"
    operation_id = "Comments_createNewTaskComment"
    action_identifier = "/task/{task_id}/comment_post"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {
        "comment_text": {"__alias": "comment_text"},
        "assignee": {"__alias": "assignee"},
        "notify_all": {"__alias": "notify_all"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteATimeEntryRequest(BaseModel):
    """Request schema for `DeleteATimeEntry`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    timer_id: int = Field(
        ...,
        alias="timer_id",
        description="Array of timer ids to delete separated by commas",
    )


class DeleteATimeEntryResponse(BaseModel):
    """Response schema for `DeleteATimeEntry`"""

    data: t.Dict[str, t.Any]


class DeleteATimeEntry(OpenAPIAction):
    """Delete a time entry from a Workspace."""

    _tags = ["Time Tracking"]
    _display_name = "delete_a_time_entry"
    _request_schema = DeleteATimeEntryRequest
    _response_schema = DeleteATimeEntryResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/{timer_id}"
    method = "delete"
    operation_id = "TimeTracking_removeEntry"
    action_identifier = "/team/{team_id}/time_entries/{timer_id}_delete"

    path_params = {"team_id": "team_id", "timer_id": "timer_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
Http client implementation for Composio SDK
"""

import typing as t

from requests import ReadTimeout
from requests import Session as SyncSession

from composio.__version__ import __version__
from composio.utils import logging


DEFAULT_RUNTIME = "composio"
SOURCE_HEADER = "python_sdk"
DEFAULT_REQUEST_TIMEOUT = 60.0


class HttpClient(SyncSession, logging.WithLogger):
    """HTTP client for Composio"""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        runtime: t.Optional[str] = None,
        timeout: t.Optional[float] = None,
    ) -> None:
        """
        Initialize client channel for Composio API

        :param base_url: Base URL for Composio API
        :param api_key: API key for Composio API
        :param runtime: Runtime specifier
        :param timeout: Request timeout
        """
        SyncSession.__init__(self)
        logging.WithLogger.__init__(self)
        self.base_url = base_url
        self.headers.update(
            {
                "x-api-key": api_key,
                "x-source": SOURCE_HEADER,
                "x-runtime": runtime or DEFAULT_RUNTIME,
                "x-composio-version": __version__,
            }
        )
        self.timeout = timeout or DEFAULT_REQUEST_TIMEOUT

    def _wrap(self, method: t.Callable) -> t.Callable:
        """Wrap http request."""

        def request(url: str, **kwargs: t.Any) -> t.Any:
            """Perform HTTP request."""
            self._logger.debug(
                f"{method.__name__.upper()} {self.base_url}{url} - {kwargs}"
            )
            retries = 0
            while retries < 3:
                try:
                    return method(
                        url=f"{self.base_url}{url}",
                        timeout=self.timeout,
                        **kwargs,
                    )
                except ReadTimeout:
                    retries += 1
            raise TimeoutError("Timed out while waiting for request to complete")

        return request

    def __getattribute__(self, name: str) -> t.Any:
        if name in ("get", "post", "put", "delete", "patch"):
            return self._wrap(super().__getattribute__(name))
        return super().__getattribute__(name)



"""
Action for creating session.
"""

# pylint: disable=import-outside-toplevel

import os
import typing as t
import uuid

from pydantic import BaseModel

from composio.tools.base.local import LocalAction


class CreateSessionRequest(BaseModel):
    """Create zep session request."""

    session_id: t.Optional[str] = None


class CreateSessionResponse(BaseModel):
    """Create zep session response."""

    session_id: str


class CreateSession(LocalAction[CreateSessionRequest, CreateSessionResponse]):
    """Create zep session."""

    _tags = ["memory", "history"]

    def execute(
        self, request: CreateSessionRequest, metadata: t.Dict
    ) -> CreateSessionResponse:
        """Create a new session."""
        from zep_cloud.client import Zep

        client = Zep(api_key=os.environ.get("ZEP_API_KEY", metadata.get("api_key")))
        session_id = request.session_id or uuid.uuid4().hex
        client.memory.add_session(session_id=session_id)
        return CreateSessionResponse(session_id=session_id)



import random
import string
import typing as t


_shell_ids: t.Set[str] = set()


def random_string(n: int = 6) -> str:
    """Random string."""
    return "".join(
        random.choices(
            string.ascii_letters + string.digits,
            k=n,
        )
    )


def generate_id() -> str:
    """Generate unique ID."""
    _id = random_string()
    while _id in _shell_ids:
        _id = random_string()
    _shell_ids.add(_id)
    return _id



import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import GetWorkspaceHistory


class HistoryFetcher(LocalTool, autoload=True):
    """Local workspace tool which can maintain history across commands"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/shelltool.png"

    @classmethod
    def actions(cls) -> list[t.Type[LocalAction]]:
        return [GetWorkspaceHistory]



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class StopATimeEntryRequest(BaseModel):
    """Request schema for `StopATimeEntry`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )


class StopATimeEntryResponse(BaseModel):
    """Response schema for `StopATimeEntry`"""

    data: t.Dict[str, t.Any]


class StopATimeEntry(OpenAPIAction):
    """Stop a timer that's currently running for the authenticated user."""

    _tags = ["Time Tracking"]
    _display_name = "stop_a_time_entry"
    _request_schema = StopATimeEntryRequest
    _response_schema = StopATimeEntryResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/stop"
    method = "post"
    operation_id = "TimeTracking_stopTimeEntry"
    action_identifier = "/team/{team_id}/time_entries/stop_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class UpdateTeamRequest(BaseModel):
    """Request schema for `UpdateTeam`"""

    group_id: str = Field(
        ...,
        alias="group_id",
        description="7C73-4002-A6A9-310014852858 (string) - Team ID (user group)",
    )
    name: str = Field(
        default=None,
        alias="name",
        description="Name",
    )
    handle: str = Field(
        default=None,
        alias="handle",
        description="Handle",
    )
    members_add: t.List[int] = Field(
        default=...,
        alias="members__add",
        description="",
    )
    members_rem: t.List[int] = Field(
        default=...,
        alias="members__rem",
        description="",
    )


class UpdateTeamResponse(BaseModel):
    """Response schema for `UpdateTeam`"""

    data: t.Dict[str, t.Any]


class UpdateTeam(OpenAPIAction):
    """
    The endpoint controls user group assignments in a Workspace, requiring 'team_id'
    and 'group_id'. Adding a view-only guest as a paid member incurs prorated
    charges for additional seats.
    """

    _tags = ["Teams - User Groups"]
    _display_name = "update_team"
    _request_schema = UpdateTeamRequest
    _response_schema = UpdateTeamResponse

    url = "https://api.clickup.com/api/v2"
    path = "/group/{group_id}"
    method = "put"
    operation_id = "TeamsUserGroups_updateUserGroup"
    action_identifier = "/group/{group_id}_put"

    path_params = {"group_id": "group_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "name": {"__alias": "name"},
        "handle": {"__alias": "handle"},
        "members": {
            "__alias": "members",
            "add": {"__alias": "add"},
            "rem": {"__alias": "rem"},
        },
    }

    aliases = {"members": "3b810d6a"}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetTimeEntryHistoryRequest(BaseModel):
    """Request schema for `GetTimeEntryHistory`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    timer_id: str = Field(
        ...,
        alias="timer_id",
        description=(
            "The ID of a time entry.    This can be found using the [Get Time Entries "
            "Within a Date Range](https://clickup.com/api/clickupreference/operation/Gettimeentrieswithinadaterange/) "
            "endpoint. "
        ),
    )


class GetTimeEntryHistoryResponse(BaseModel):
    """Response schema for `GetTimeEntryHistory`"""

    data: t.Dict[str, t.Any]


class GetTimeEntryHistory(OpenAPIAction):
    """View a list of changes made to a time entry."""

    _tags = ["Time Tracking"]
    _display_name = "get_time_entry_history"
    _request_schema = GetTimeEntryHistoryRequest
    _response_schema = GetTimeEntryHistoryResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/time_entries/{timer_id}/history"
    method = "get"
    operation_id = "TimeTracking_getTimeEntryHistory"
    action_identifier = "/team/{team_id}/time_entries/{timer_id}/history_get"

    path_params = {"team_id": "team_id", "timer_id": "timer_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
E2E Tests for plugin demos and examples.
"""

import os
import subprocess
import sys
import typing as t
from pathlib import Path

import pytest


PLUGINS = Path.cwd() / "plugins"
EXAMPLES_PATH = Path.cwd() / "examples"

# Require env vars
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
COMPOSIO_API_KEY = os.environ.get("COMPOSIO_API_KEY")
JULEP_API_KEY = os.environ.get("JULEP_API_KEY")
JULEP_API_URL = os.environ.get("JULEP_API_URL")
LISTENNOTES_API_KEY = os.environ.get("LISTENNOTES_API_KEY")

# Plugin test definitions
EXAMPLES = {
    "autogen": {
        "plugin": "autogen",
        "file": PLUGINS / "autogen" / "autogen_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {
            "OPENAI_API_KEY": OPENAI_API_KEY,
            "COMPOSIO_API_KEY": COMPOSIO_API_KEY,
        },
    },
    "llamaindex": {
        "plugin": "llamaindex",
        "file": PLUGINS / "llamaindex" / "llamaindex_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {
            "OPENAI_API_KEY": OPENAI_API_KEY,
            "COMPOSIO_API_KEY": COMPOSIO_API_KEY,
        },
    },
    "local_tools": {
        "plugin": "autogen",
        "file": EXAMPLES_PATH / "miscellaneous" / "math_agent" / "autogen_math.py",
        "match": {
            "type": "stdout",
            "values": ["11962"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY},
    },
    "runtime_tools": {
        "plugin": "langchain",
        "file": EXAMPLES_PATH / "miscellaneous" / "runtime_tools" / "langchain_math.py",
        "match": {
            "type": "stdout",
            "values": ["2645406630"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY},
    },
    "crew_ai": {
        "plugin": "crew_ai",
        "file": PLUGINS / "crew_ai" / "crewai_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {
            "OPENAI_API_KEY": OPENAI_API_KEY,
            "COMPOSIO_API_KEY": COMPOSIO_API_KEY,
        },
    },
    # TOFIX(@kaave): httpcore.UnsupportedProtocol: Request URL is missing an 'http://' or 'https://' protocol.
    # "julep": {
    #     "plugin": "julep",
    #     "file": PLUGINS / "julep" / "julep_demo.py",
    #     "match": {
    #         "type": "stdout",
    #         "values": ["finish_reason=<ChatResponseFinishReason.STOP: 'stop'>"],
    #     },
    #     "env": {
    #         "OPENAI_API_KEY": OPENAI_API_KEY,
    #         "COMPOSIO_API_KEY": COMPOSIO_API_KEY,
    #         "JULEP_API_KEY": JULEP_API_KEY,
    #         "JULEP_API_URL": JULEP_API_URL,
    #     },
    # },
    "langchain": {
        "plugin": "langchain",
        "file": PLUGINS / "langchain" / "langchain_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    },
    "langgraph": {
        "plugin": "langgraph",
        "file": PLUGINS / "langgraph" / "langgraph_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    },
    "openai": {
        "plugin": "openai",
        "file": PLUGINS / "openai" / "openai_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    },
    "lyzr": {
        "plugin": "lyzr",
        "file": PLUGINS / "lyzr" / "lyzr_demo.py",
        "match": {
            "type": "stdout",
            "values": ["Action executed successfully"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    },
    "upload_file": {
        "plugin": "crew_ai",
        "file": EXAMPLES_PATH / "miscellaneous" / "attachment" / "send_attachment.py",
        "match": {
            "type": "stdout",
            "values": ["'labelIds': ['SENT']"],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    },
    "download_file": {
        "plugin": "crew_ai",
        "file": "run_issue.py",
        "match": {
            "type": "stdout",
            "values": ["composio_output/CODEINTERPRETER_GET_FILE_CMD_default_", ""],
        },
        "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
        "cwd": EXAMPLES_PATH
        / "quickstarters"
        / "sql_agent"
        / "sql_agent_plotter_crewai",
    },
    "multi_entity_api_key": {
        "plugin": "langchain",
        "file": EXAMPLES_PATH / "miscellaneous" / "multi_entity.py",
        "match": {
            "type": "stdout",
            "values": [
                "Invoking: `LISTENNOTES_FETCH_A_LIST_OF_SUPPORTED_LANGUAGES_FOR_PODCASTS`",
                "Any language",
                "Abkhazian",
                "Arabic",
            ],
        },
        "env": {
            "OPENAI_API_KEY": OPENAI_API_KEY,
            "COMPOSIO_API_KEY": COMPOSIO_API_KEY,
            "LISTENNOTES_API_KEY": LISTENNOTES_API_KEY,
        },
    },
    # "praisonai": {
    #     "plugin": "praisonai",
    #     "file": PLUGINS / "praisonai" / "praisonai_demo.py",
    #     "match": {
    #         "type": "stdout",
    #         "values": [
    #             "Action executed successfully"
    #         ],
    #     },
    #     "env": {"OPENAI_API_KEY": OPENAI_API_KEY, "COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    # },
    # TODO(@kaavee): Add Anthropic API key
    # "claude": {
    #     "plugin": "claude",
    #     "file": PLUGINS / "claude" / "claude_demo.py",
    #     "match": {
    #         "type": "stdout",
    #         "values": ["Action executed successfully"],
    #     },
    #     "env": {"COMPOSIO_API_KEY": COMPOSIO_API_KEY},
    # },
    # TODO: Add camel
}


@pytest.mark.skipif(
    condition=os.environ.get("CI", "false") == "true",
    reason="Testing in CI will lead to too much LLM API usage",
)
@pytest.mark.parametrize("example_name, example", EXAMPLES.items())
def test_example(
    example_name: str, example: dict  # pylint: disable=unused-argument
) -> None:
    """Test an example with given environment."""
    plugin_to_test = os.getenv("PLUGIN_TO_TEST")
    if plugin_to_test is not None and plugin_to_test != example["plugin"]:
        pytest.skip(f"Skipping {example['plugin']}")

    for key, val in example["env"].items():
        assert (
            val is not None
        ), f"Please provide value for `{key}` for testing `{example['file']}`"

    cwd = example.get("cwd", None)
    proc = subprocess.Popen(  # pylint: disable=consider-using-with
        args=[sys.executable, example["file"]],
        # TODO(@angryblade): Sanitize the env before running the process.
        env={**os.environ, **example["env"]},
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
    )

    # Wait for 2 minutes for example to run
    proc.wait(timeout=180)

    # Check if process exited with success
    assert proc.returncode == 0, (
        t.cast(t.IO[bytes], proc.stderr).read().decode(encoding="utf-8")
    )

    # Validate output
    output = (
        t.cast(
            t.IO[bytes],
            (proc.stdout if example["match"]["type"] == "stdout" else proc.stderr),
        )
        .read()
        .decode(encoding="utf-8")
    )
    for match in example["match"]["values"]:
        assert match in output



import typing as t

from pydantic import BaseModel

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetAuthorizedUserRequest(BaseModel):
    """Request schema for `GetAuthorizedUser`"""


class GetAuthorizedUserResponse(BaseModel):
    """Response schema for `GetAuthorizedUser`"""

    data: t.Dict[str, t.Any]


class GetAuthorizedUser(OpenAPIAction):
    """View the details of the authenticated user's ClickUp account."""

    _tags = ["Authorization"]
    _display_name = "get_authorized_user"
    _request_schema = GetAuthorizedUserRequest
    _response_schema = GetAuthorizedUserResponse

    url = "https://api.clickup.com/api/v2"
    path = "/user"
    method = "get"
    operation_id = "Authorization_viewAccountDetails"
    action_identifier = "/user_get"

    path_params = {}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class authorization_view_account_details(OpenAPIAction):
    """
    View the details of the authenticated user's ClickUp account.<<DEPRECATED
    use get_authorized_user>>
    """

    _tags = ["Authorization"]
    _display_name = "authorization_view_account_details"
    _request_schema = GetAuthorizedUserRequest
    _response_schema = GetAuthorizedUserResponse

    url = "https://api.clickup.com/api/v2"
    path = "/user"
    method = "get"
    operation_id = "Authorization_viewAccountDetails"
    action_identifier = "/user_get"

    path_params = {}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetGoalRequest(BaseModel):
    """Request schema for `GetGoal`"""

    goal_id: str = Field(
        ...,
        alias="goal_id",
        description="900e-462d-a849-4a216b06d930 (uuid)",
    )


class GetGoalResponse(BaseModel):
    """Response schema for `GetGoal`"""

    data: t.Dict[str, t.Any]


class GetGoal(OpenAPIAction):
    """View the details of a Goal including its Targets."""

    _tags = ["Goals"]
    _display_name = "get_goal"
    _request_schema = GetGoalRequest
    _response_schema = GetGoalResponse

    url = "https://api.clickup.com/api/v2"
    path = "/goal/{goal_id}"
    method = "get"
    operation_id = "Goals_getDetails"
    action_identifier = "/goal/{goal_id}_get"

    path_params = {"goal_id": "goal_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class RemoveUserFromWorkspaceRequest(BaseModel):
    """Request schema for `RemoveUserFromWorkspace`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    user_id: int = Field(
        ...,
        alias="user_id",
        description="",
    )


class RemoveUserFromWorkspaceResponse(BaseModel):
    """Response schema for `RemoveUserFromWorkspace`"""

    data: t.Dict[str, t.Any]


class RemoveUserFromWorkspace(OpenAPIAction):
    """
    Deactivate a user from a Workspace.    ***Note:** This endpoint is only
    available to Workspaces on our [Enterprise Plan](https://clickup.com/pricing).*
    """

    _tags = ["Users"]
    _display_name = "remove_user_from_workspace"
    _request_schema = RemoveUserFromWorkspaceRequest
    _response_schema = RemoveUserFromWorkspaceResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/user/{user_id}"
    method = "delete"
    operation_id = "Users_deactivateFromWorkspace"
    action_identifier = "/team/{team_id}/user/{user_id}_delete"

    path_params = {"team_id": "team_id", "user_id": "user_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class DeleteCommentRequest(BaseModel):
    """Request schema for `DeleteComment`"""

    comment_id: int = Field(
        ...,
        alias="comment_id",
        description="",
    )


class DeleteCommentResponse(BaseModel):
    """Response schema for `DeleteComment`"""

    data: t.Dict[str, t.Any]


class DeleteComment(OpenAPIAction):
    """Delete a task comment."""

    _tags = ["Comments"]
    _display_name = "delete_comment"
    _request_schema = DeleteCommentRequest
    _response_schema = DeleteCommentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/comment/{comment_id}"
    method = "delete"
    operation_id = "Comments_deleteTaskComment"
    action_identifier = "/comment/{comment_id}_delete"

    path_params = {"comment_id": "comment_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetWorkspaceSeatsRequest(BaseModel):
    """Request schema for `GetWorkspaceSeats`"""

    team_id: str = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )


class GetWorkspaceSeatsResponse(BaseModel):
    """Response schema for `GetWorkspaceSeats`"""

    data: t.Dict[str, t.Any]


class GetWorkspaceSeats(OpenAPIAction):
    """
    View the used, total, and available member and guest seats for a Workspace.
    """

    _tags = ["Teams - Workspaces"]
    _display_name = "get_workspace_seats"
    _request_schema = GetWorkspaceSeatsRequest
    _response_schema = GetWorkspaceSeatsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/seats"
    method = "get"
    operation_id = "TeamsWorkspaces_getWorkspaceSeats"
    action_identifier = "/team/{team_id}/seats_get"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class teams_work_spaces_get_work_space_seats(OpenAPIAction):
    """
    View the used, total, and available member and guest seats for a Workspace.<<DEPRECATED
    use get_workspace_seats>>
    """

    _tags = ["Teams - Workspaces"]
    _display_name = "teams_work_spaces_get_work_space_seats"
    _request_schema = GetWorkspaceSeatsRequest
    _response_schema = GetWorkspaceSeatsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/seats"
    method = "get"
    operation_id = "TeamsWorkspaces_getWorkspaceSeats"
    action_identifier = "/team/{team_id}/seats_get"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""FlyIO Client."""

import os
import time
import typing as t
import uuid

import requests
import typing_extensions as te

from composio.tools.env.constants import DEFAULT_IMAGE
from composio.utils.logging import WithLogger


try:
    import gql
    import gql.transport
    from gql.transport.requests import RequestsHTTPTransport

    FLYIO_DEPENDENCIES_INSTALLED = True
except ImportError:
    RequestsHTTPTransport = t.Any
    FLYIO_DEPENDENCIES_INSTALLED = False


FLY_API = "https://api.machines.dev"
FLY_GRAPHQL_API = "https://api.fly.io/graphql"
API_VERSION = "/v1"
BASE_URL = FLY_API + API_VERSION


ALLOCATE_IP_QUERY = """mutation {
  allocateIpAddress(input: { appId: "{app_name}", type: shared_v4 }) {
    ipAddress {
      address
      type
    }
    clientMutationId
  }
}
"""

GET_IP_REQUEST = """query {
  app(name:"{app_name}"){
    sharedIpAddress
  }
}
"""


RELEASE_IP_REQUEST = """mutation {
  releaseIpAddress(input: { appId: "{app_name}", ip: "{ip}" }) {
    app {
      id
    }
    clientMutationId
  }
}
"""

ENV_FLY_API_TOKEN = "FLY_API_TOKEN"


TOOLING_ERVICE = {
    "ports": [
        {"port": 8000, "handlers": ["tls", "http"]},
    ],
    "protocol": "tcp",
    "internal_port": 8000,
}


class ExternalPortConfig(te.TypedDict):
    """External port config."""

    port: int
    """External port to map (Port 8000 is reserved for the tooling server)."""

    handlers: t.List[str]
    """List of protocol handlers (`http` will be used if not provided)."""


class PortRequest(te.TypedDict):
    """
    Port request

    Read more at: https://fly.io/docs/machines/api/machines-resource/#create-a-machine-with-services
    """

    ports: t.List[ExternalPortConfig]
    """List of public port configurations (Port 8000 is reserved for the tooling server)."""

    internal_port: int
    """Internal port number (Port 8000 is reserved for the tooling server)."""

    protocol: t.Literal["tcp", "udp"]
    """List of protocol handlers."""


class FlyIO(WithLogger):
    """FlyIO client."""

    machine: str

    def __init__(
        self,
        access_token: str,
        image: t.Optional[str] = None,
        flyio_token: t.Optional[str] = None,
        environment: t.Optional[t.Dict] = None,
        ports: t.Optional[t.List[PortRequest]] = None,
    ) -> None:
        """Initialize FlyIO client."""
        super().__init__()
        flyio_token = flyio_token or os.environ.get(ENV_FLY_API_TOKEN)
        if flyio_token is None:
            raise ValueError(
                "FlyIO API Key is required for using FlyIO workspace, "
                f"You can export it as `{ENV_FLY_API_TOKEN}`"
            )

        self.ports = ports or []
        self.environment = environment or {}
        self.flyio_token = flyio_token
        self.access_token = access_token
        self.image = image or DEFAULT_IMAGE
        self.app_name = f"composio-{uuid.uuid4().hex.replace('-', '')}"
        self.url = f"https://{self.app_name}.fly.dev:8000/api"
        self.gql = gql.Client(
            transport=RequestsHTTPTransport(
                url=FLY_GRAPHQL_API,
                headers={
                    "Authorization": f"Bearer {self.flyio_token}",
                },
            )
        )

    def _request(
        self,
        method: str,
        endpoint: str,
        json: t.Optional[dict] = None,
        timeout: t.Optional[float] = 15.0,
    ) -> requests.Response:
        """Make request to machines API."""
        return requests.request(
            method=method,
            url=f"{BASE_URL}{endpoint}",
            json=json,
            headers={
                "Authorization": f"Bearer {self.flyio_token}",
            },
            timeout=timeout,
        )

    def _create_app(self) -> None:
        self._request(
            method="post",
            endpoint="/apps",
            json={
                "app_name": self.app_name,
                "org_slug": "personal",
            },
        ).json()

    def _delete_app(self) -> None:
        self._request(
            method="delete",
            endpoint=f"/apps/{self.app_name}",
        ).content.decode()

    def _allocate_ip(self) -> None:
        self.gql.execute(
            document=gql.gql(
                request_string=ALLOCATE_IP_QUERY.replace(
                    "{app_name}",
                    self.app_name,
                )
            ),
        )

    def _get_ip(self) -> str:
        return (
            self.gql.execute(
                document=gql.gql(
                    request_string=GET_IP_REQUEST.replace(
                        "{app_name}",
                        self.app_name,
                    )
                ),
            )
            .get("app")
            .get("sharedIpAddress")  # type: ignore
        )

    def _release_ip(self) -> None:
        self.gql.execute(
            document=gql.gql(
                request_string=RELEASE_IP_REQUEST.replace(
                    "{app_name}",
                    self.app_name,
                ).replace(
                    "{ip}",
                    self._get_ip(),
                )
            ),
        )

    def _wait_for_machine(self) -> None:
        """Wait for machine to get started."""
        deadline = time.time() + float(os.environ.get("WORKSPACE_WAIT_TIMEOUT", 60.0))
        while time.time() < deadline:
            try:
                requests.get(
                    self.url,
                    headers={"x-api-key": self.access_token},
                    timeout=30.0,
                ).content.decode()
                return
            except Exception:  # pylint: disable=broad-except
                time.sleep(1)

    def _create_machine(self) -> None:
        self.machine = (
            self._request(
                method="post",
                endpoint=f"/apps/{self.app_name}/machines",
                json={
                    "config": {
                        "image": self.image,
                        "env": self.environment,
                        "services": [TOOLING_ERVICE, *self.ports],
                        "guest": {
                            "cpu_kind": "shared",
                            "cpus": 1,
                            "memory_mb": 512,
                        },
                    }
                },
            )
            .json()
            .get("id")
        )

    def _stop_machine(self) -> None:
        self._request(
            method="post",
            endpoint=f"/apps/{self.app_name}/machines/{self.machine}/stop",
        ).json()

    def _delete_machine(self) -> None:
        self._request(
            method="delete",
            endpoint=f"/apps/{self.app_name}/machines/{self.machine}?force=true",
        ).json()

    def setup(self) -> None:
        """Setup FlyIO machine."""
        self.logger.debug(f"Creating app with name `{self.app_name}`")
        self._create_app()
        self.logger.debug("Allocating IP to the app")
        self._allocate_ip()
        self.logger.debug(f"Allocated IP `{self._get_ip()}` to app `{self.app_name}`")
        self._create_machine()
        self.logger.debug("Waiting for machine to start")
        self._wait_for_machine()
        self.logger.debug(f"Created machine {self.machine}")

    def teardown(self) -> None:
        """Teardown FlyIO machine."""
        self.logger.debug(f"Stopping machine `{self.machine}`")
        self._stop_machine()
        self.logger.debug(f"Deleting machine `{self.machine}`")
        self._delete_machine()
        self.logger.debug(f"Releasing IP for `{self.app_name}`")
        self._release_ip()
        self.logger.debug(f"Deleting app `{self.app_name}`")
        self._delete_app()



from functools import wraps
from typing import Callable, Dict, TypeVar

from pydantic import BaseModel, Field


class BaseFileRequest(BaseModel):
    file_manager_id: str = Field(
        default="",
        description="ID of the file manager where the file will be opened, if not "
        "provided the recent file manager will be used to execute the action",
    )


class BaseFileResponse(BaseModel):
    error: str = Field(
        default="",
        description="Error message if the action failed",
    )
    current_working_directory: str = Field(
        default="",
        description="Current working directory of the file manager.",
    )


T = TypeVar("T", bound=BaseFileRequest)
R = TypeVar("R", bound=BaseFileResponse)


# TOFIX: Override `Filetool.execute` and get rid of this decorator
def include_cwd(func: Callable[[T, Dict], R]) -> Callable[[T, Dict], R]:
    @wraps(func)
    def wrapper(self, request: T, metadata: Dict) -> R:
        response = func(self, request, metadata)  # type: ignore
        response.current_working_directory = str(
            self.filemanagers.get(request.file_manager_id).working_dir
        )
        return response

    return wrapper  # type: ignore



from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


# pylint: disable=import-outside-toplevel
class RagToolAddRequest(BaseModel):
    content: str = Field(
        ...,
        description="Content to add to the knowledge base",
        json_schema_extra={"file_readable": True},
    )


class RagToolAddResponse(BaseModel):
    status: str = Field(..., description="Status of the addition to the knowledge base")


class AddContentToRagTool(LocalAction[RagToolAddRequest, RagToolAddResponse]):
    """Tool for adding content to the knowledge base"""

    _tags = ["Knowledge Base"]

    def execute(self, request: RagToolAddRequest, metadata: Dict) -> RagToolAddResponse:
        """Add content to the knowledge base"""
        try:
            from embedchain import App
        except ImportError as e:
            raise ImportError(f"Failed to import App from embedchain: {e}") from e

        App().add(request.content)
        return RagToolAddResponse(status="Content added successfully")



import typing as t

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class SearchWordRequest(BaseFileRequest):
    """Request to search for a word in files."""

    word: str = Field(..., description="The term to search for")
    pattern: str = Field(
        default=None,
        description="""The file, directory, or glob pattern to search in.
        If not provided, searches in the current working directory.
        Examples:
        - "*.py" : Search in all Python files in the current directory
        - "src/*.txt" : Search in all text files in the 'src' directory
        - "**/*.md" : Search in all Markdown files in the current directory and all subdirectories
        - "/path/to/specific/file.js" : Search in a specific file
        - "/path/to/directory" : Search in all files in a specific directory""",
    )
    recursive: bool = Field(
        default=True, description="If True, search recursively in subdirectories"
    )
    case_insensitive: bool = Field(
        default=True, description="If True, perform case-insensitive search"
    )
    exclude: t.List[str] = Field(
        default=None,
        description="List of directories to exclude from the search",
    )


class SearchWordResponse(BaseFileResponse):
    """Response to search for a word in files."""

    results: t.Dict[str, t.List[t.Tuple[int, str]]] = Field(
        default={},
        description="A dictionary with file paths as keys and lists of (line number, line content) tuples as values",
    )
    message: str = Field(default="", description="Message to display to the user")
    error: str = Field(default="", description="Error message if any")


class SearchWord(LocalAction[SearchWordRequest, SearchWordResponse]):
    """
    - Search for a specific word or phrase across multiple files
    in your workspace by specifying a pattern.
    You can specify a pattern to narrow down the search to specific
    files, directories, or file types.
    The search returns the line numbers and content of the lines
    where the word is found.

    Usage examples:
    1. Search for 'TODO' in all Python files:
       pattern: "*.py", word: "TODO"
    2. Search for 'bug' in all files in the 'src' directory:
       pattern: "src/*", word: "bug"
    3. Search for 'FIXME' in all JavaScript files in any subdirectory:
       pattern: "**/*.js", word: "FIXME"
    4. Search for 'important' in a specific file:
       pattern: "/path/to/specific/file.txt", word: "important"

    Note: The search will skip binary files and hidden files (those starting with a dot).

    Raises:
        - ValueError: If the word to search for is empty.
        - FileNotFoundError: If the specified pattern doesn't match any files.
    """

    @include_cwd  # type: ignore
    def execute(
        self, request: SearchWordRequest, metadata: t.Dict
    ) -> SearchWordResponse:
        try:
            results = self.filemanagers.get(request.file_manager_id).grep(
                word=request.word,
                pattern=request.pattern,
                recursive=request.recursive,
                case_insensitive=request.case_insensitive,
                exclude=request.exclude,  # type: ignore
            )
            num_files: int = len(results)
            if num_files > 100:
                return SearchWordResponse(
                    results=dict(list(results.items())[:100]),
                    message=(
                        f'Warning: More than 100 files matched for "{request.word}" '
                        f'in "{request.pattern}". Sending the first 100 results. '
                        "Consider narrowing your search."
                    ),
                )
            if num_files == 0:
                return SearchWordResponse(
                    results={},
                    message=f'No files matched for "{request.word}" in {request.pattern}".',
                )
            return SearchWordResponse(results=results)
        except ValueError as e:
            return SearchWordResponse(error=f"Invalid search parameters: {str(e)}")
        except FileNotFoundError as e:
            return SearchWordResponse(
                error=f"No files found matching the pattern: {str(e)}"
            )
        except PermissionError as e:
            return SearchWordResponse(error=f"Permission denied: {str(e)}")
        except IOError as e:
            return SearchWordResponse(error=f"Error reading files: {str(e)}")



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetTaskTemplatesRequest(BaseModel):
    """Request schema for `GetTaskTemplates`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    page: int = Field(
        ...,
        alias="page",
        description="",
    )


class GetTaskTemplatesResponse(BaseModel):
    """Response schema for `GetTaskTemplates`"""

    data: t.Dict[str, t.Any]


class GetTaskTemplates(OpenAPIAction):
    """View the task templates available in a Workspace."""

    _tags = ["Task Templates"]
    _display_name = "get_task_templates"
    _request_schema = GetTaskTemplatesRequest
    _response_schema = GetTaskTemplatesResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/taskTemplate"
    method = "get"
    operation_id = "TaskTemplates_getTemplates"
    action_identifier = "/team/{team_id}/taskTemplate_get"

    path_params = {"team_id": "team_id"}
    query_params = {"page": "page"}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from pydantic import BaseModel

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetAuthorizedTeamsWorkspacesRequest(BaseModel):
    """Request schema for `GetAuthorizedTeamsWorkspaces`"""


class GetAuthorizedTeamsWorkspacesResponse(BaseModel):
    """Response schema for `GetAuthorizedTeamsWorkspaces`"""

    data: t.Dict[str, t.Any]


class GetAuthorizedTeamsWorkspaces(OpenAPIAction):
    """View the Workspaces available to the authenticated user."""

    _tags = ["Authorization", "Teams - Workspaces"]
    _display_name = "get_authorized_teams_workspaces"
    _request_schema = GetAuthorizedTeamsWorkspacesRequest
    _response_schema = GetAuthorizedTeamsWorkspacesResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team"
    method = "get"
    operation_id = "Authorization_getWorkspaceList"
    action_identifier = "/team_get"

    path_params = {}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class authorization_get_work_space_list(OpenAPIAction):
    """
    View the Workspaces available to the authenticated user.<<DEPRECATED use
    get_authorized_teams_workspaces>>
    """

    _tags = ["Authorization", "Teams - Workspaces"]
    _display_name = "authorization_get_work_space_list"
    _request_schema = GetAuthorizedTeamsWorkspacesRequest
    _response_schema = GetAuthorizedTeamsWorkspacesResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team"
    method = "get"
    operation_id = "Authorization_getWorkspaceList"
    action_identifier = "/team_get"

    path_params = {}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



from typing import Dict

from pydantic import Field

from composio.tools.base.exceptions import ExecutionFailed
from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
)


class RenameFileRequest(BaseFileRequest):
    """Request to rename a file."""

    old_file_path: str = Field(
        ...,
        description="Old file path to rename. This is a relative path to the current directory",
    )
    new_file_path: str = Field(
        ...,
        description="New file path to rename. This is a relative path to the current directory",
    )


class RenameFileResponse(BaseFileResponse):
    """Response to rename a file."""

    message: str = Field(default="", description="Message to display to the user")
    error: str = Field(default="", description="Error message if any")


class RenameFile(LocalAction[RenameFileRequest, RenameFileResponse]):
    """
    Renames a file based on the provided file path,

    Can result in:
    - ValueError: If old_file_path / new_file_path is not a string or if the file/directory does not exist.
    - FileExistsError: If the new_file_path file/directory path already exists.
    - FileNotFoundError: If the old_file_path file/directory does not exist.
    - PermissionError: If the user doesn't have permission to rename the file/directory.
    """

    display_name = "Rename a file"

    def execute(self, request: RenameFileRequest, metadata: Dict) -> RenameFileResponse:
        is_success = self.filemanagers.get(
            request.file_manager_id,
        ).rename(
            request.old_file_path,
            request.new_file_path,
        )
        if not is_success:
            raise ExecutionFailed("Failed to rename the file.")
        return RenameFileResponse(
            message="File renamed successfully.",
        )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetUserRequest(BaseModel):
    """Request schema for `GetUser`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    user_id: int = Field(
        ...,
        alias="user_id",
        description="",
    )


class GetUserResponse(BaseModel):
    """Response schema for `GetUser`"""

    data: t.Dict[str, t.Any]


class GetUser(OpenAPIAction):
    """
    View information about a user in a Workspace.    ***Note:** This endpoint
    is only available to Workspaces on our [Enterprise Plan](https://clickup.com/pricing).*
    """

    _tags = ["Users"]
    _display_name = "get_user"
    _request_schema = GetUserRequest
    _response_schema = GetUserResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/user/{user_id}"
    method = "get"
    operation_id = "Users_getUserDetails"
    action_identifier = "/team/{team_id}/user/{user_id}_get"

    path_params = {"team_id": "team_id", "user_id": "user_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""Spawn a process."""

import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


# pylint: disable=consider-using-with,unspecified-encoding


class SpawnRequest(BaseModel):
    """Execute command request."""

    cmd: str = Field(
        ...,
        description="Command to be executed.",
        examples=[
            "/bin/python /home/user/server.py",
            "node /home/user/server.js",
            "yarn start",
        ],
    )
    working_dir: str = Field(
        None,
        description=(
            "Directory where this command should be executed, "
            "if not provided the current directory will be used"
        ),
        examples=[
            "/home/user",
            "./",
        ],
    )


class SpawnResponse(BaseModel):
    """Shell execution response."""

    stdout: str = Field(
        ...,
        description="Path to the file containing the stdout stream",
    )
    stderr: str = Field(
        ...,
        description="Path to the file containing the stderr stream.",
    )
    pid: str = Field(
        ...,
        description="Path to the file containing the process ID for the spawned process",
    )


class SpawnProcess(LocalAction[SpawnRequest, SpawnResponse]):
    """
    Spawn a process.

    Use this action to launch processes on background, for example launch a
    python process using

    cmd: python path/to/script.py
    """

    _tags = ["workspace", "shell"]

    def execute(self, request: SpawnRequest, metadata: Dict) -> SpawnResponse:
        """Execute a shell command."""
        _cmd, *args = request.cmd.split(" ")
        cmd = shutil.which(cmd=_cmd)  # type: ignore
        if cmd is None:
            raise ValueError(f"Command `{_cmd}` not found!")

        tempdir = tempfile.TemporaryDirectory()
        stdout = Path(tempdir.name, "stdout.txt")
        stderr = Path(tempdir.name, "stderr.txt")
        process = subprocess.Popen(
            args=[cmd, *args],
            start_new_session=True,
            stdout=stdout.open("w+"),
            stderr=stderr.open("w+"),
            cwd=str(request.working_dir or Path.cwd()),
        )
        pid = Path(tempdir.name, "pid.txt")
        pid.write_text(str(process.pid))
        return SpawnResponse(stdout=str(stdout), stderr=str(stderr), pid=str(pid))



"""
Local account storage helpers.
"""

import typing as t

from composio.storage.base import LocalStorage


class UserData(LocalStorage):
    """
    Local user data storage.
    """

    api_key: t.Optional[str] = None
    """
    API key for Composio API server
    """



"""Tool abstractions."""

import inspect
import typing as t
from abc import abstractmethod

import inflection
import typing_extensions as te
from pydantic import BaseModel, Field

from composio import Composio
from composio.client.collections import CustomAuthParameter
from composio.client.enums.base import ActionData, SentinalObject, add_runtime_action
from composio.client.exceptions import ComposioClientError
from composio.exceptions import ComposioSDKError
from composio.tools.base.abs import (
    Action,
    ActionRequest,
    ActionResponse,
    ToolBuilder,
    tool_registry,
)
from composio.tools.base.local import LocalToolMixin
from composio.tools.env.host.shell import Shell
from composio.tools.env.host.workspace import Browsers, FileManagers, Shells


if t.TYPE_CHECKING:
    from composio.tools.toolset import ComposioToolSet


class InvalidRuntimeAction(ComposioSDKError):
    """Raise when invalid action definition is found"""


class FileModel(BaseModel):
    name: str = Field(
        ...,
        description="File name, contains extension to indetify the file type",
    )
    content: bytes = Field(
        ...,
        description="File content in base64",
    )


class ArgSpec(BaseModel):
    """Argument specification."""

    description: str
    """Description of the argument variable."""

    default: t.Any = None
    """Default value"""


class RuntimeAction(  # pylint: disable=abstract-method
    SentinalObject,
    Action[ActionRequest, ActionResponse],
    abs=True,
):
    """Local action abstraction."""

    _shells: t.Callable[[], Shells]
    _browsers: t.Callable[[], Browsers]
    _filemanagers: t.Callable[[], FileManagers]

    @property
    def shells(self) -> Shells:
        return self._shells()

    @property
    def browsers(self) -> Browsers:
        return self._browsers()

    @property
    def filemanagers(self) -> FileManagers:
        return self._filemanagers()


class RuntimeToolMeta(type):
    """Tool metaclass."""

    def __init__(  # pylint: disable=self-cls-assignment,unused-argument
        cls,
        name: str,
        bases: t.Tuple,
        dict_: t.Dict,
        autoload: bool = False,
    ) -> None:
        """Initialize action class."""
        if name == "RuntimeTool":
            return

        cls = t.cast(t.Type[RuntimeTool], cls)
        ToolBuilder.validate(obj=cls, name=name, methods=("actions",))
        ToolBuilder.set_metadata(obj=cls)
        ToolBuilder.setup_children(obj=cls)

        if autoload:
            cls.register()


class RuntimeTool(LocalToolMixin, metaclass=RuntimeToolMeta):
    """Local tool class."""

    gid = "runtime"
    """Group ID for this tool."""

    @classmethod
    @abstractmethod
    def actions(cls) -> t.List[t.Type[RuntimeAction]]:
        """Get collection of actions for the tool."""


def _create_tool_class(
    name: str,
    actions: t.List[t.Type[RuntimeAction]],
) -> t.Type[RuntimeTool]:
    """Create runtime tool class."""

    class _Tool:
        gid = "runtime"

        @classmethod
        def actions(cls) -> t.List[type[RuntimeAction]]:
            return actions

    _Tool.__doc__ = f"{name.title()} tool."

    return type(inflection.camelize(name), (_Tool, RuntimeTool), dict(_Tool.__dict__))


def _wrap(
    f: t.Callable,
    toolname: str,
    tags: t.List,
    file: str,
    request_schema: t.Type[BaseModel],
    response_schema: t.Type[BaseModel],
    runs_on_shell: bool = False,
    requires: t.Optional[t.List[str]] = None,
) -> t.Type[RuntimeAction]:
    """Wrap action class with given params."""

    _file = file
    _requires = requires

    class WrappedAction(RuntimeAction[request_schema, response_schema]):  # type: ignore
        """Wrapped action class."""

        _tags: t.List[str] = tags

        tool = toolname
        name = f.__name__
        enum = f.__name__.upper()
        display_name = f.__name__

        file = _file
        requires = _requires
        run_on_shell: bool = runs_on_shell

        data = ActionData(
            name=f.__name__,
            app=toolname.upper(),
            tags=tags,
            no_auth=True,
            is_local=True,
            is_runtime=True,
            shell=run_on_shell,
        )

        def execute(self, request: t.Any, metadata: dict) -> t.Any:
            return f(request, metadata)

    cls = t.cast(
        t.Type[WrappedAction],
        type(inflection.camelize(f.__name__), (WrappedAction,), {}),
    )
    cls.__doc__ = f.__doc__
    cls.description = f.__doc__  # type: ignore

    existing_actions = []
    toolname = toolname.upper()
    if toolname in tool_registry["runtime"]:
        existing_actions = tool_registry["runtime"][toolname].actions()
    tool = _create_tool_class(name=toolname, actions=[cls, *existing_actions])  # type: ignore
    tool_registry["runtime"][toolname] = tool()
    add_runtime_action(cls.enum, cls.data)
    return cls


def _is_simple_action(argspec: inspect.FullArgSpec) -> bool:
    """Check if the action is defined with `request_data` and `metadata`"""
    if "request" not in argspec.args and "metadata" not in argspec.args:
        return False

    if not issubclass(argspec.annotations["request"], BaseModel):
        raise ValueError("`request_data` needs to be a `pydantic.BaseModel` object")

    if not issubclass(argspec.annotations["return"], BaseModel):
        raise ValueError("Return type needs to be a `pydantic.BaseModel` object")

    return True


def _parse_raw_type(argument: str, annotation: t.Type) -> t.Tuple[t.Type, str]:
    """Parse for raw type."""
    return annotation, " ".join(argument.split("_")).title()


def _parse_annotated_type(
    argument: str, annotation: t.Type
) -> t.Tuple[t.Type, str, t.Any]:
    """Parse for raw type."""
    annottype, *annotspec = t.get_args(annotation)
    if len(annotspec) == 1 and isinstance(annotspec[0], ArgSpec):
        description = annotspec[0].description
        default = annotspec[0].default
    elif len(annotspec) == 1 and isinstance(annotspec[0], str):
        description = annotspec[0]
        default = None
    elif len(annotspec) == 2 and isinstance(annotspec[0], str):
        description = annotspec[0]
        default = annotspec[1]
    else:
        raise ValueError(
            f"Invalid type annotation for argument {argument}: {annotation}"
        )
    return annottype, description, default


def _parse_docstring(
    docstr: str,
) -> t.Tuple[str, t.Dict[str, str], t.Optional[t.Tuple[str, str]]]:
    """Parse docstring for descriptions."""
    header, *descriptions = docstr.lstrip().rstrip().split("\n")
    params = {}
    returns = None
    for description in descriptions:
        if not description:
            continue

        if ":param" in description:
            param, description = description.replace(":param ", "").split(":")
            params[param.lstrip().rstrip()] = description.lstrip().rstrip()

        if ":return" in description:
            param, description = description.replace(":return ", "").split(":")
            returns = (param.lstrip().strip(), description.lstrip().rstrip())

    return header, params, returns


def _get_auth_params(app: str, entity_id: str) -> t.Optional[t.Dict]:
    try:
        client = Composio.get_latest()
        connection_params = client.connected_accounts.get(
            connection_id=client.get_entity(entity_id).get_connection(app=app).id
        ).connectionParams
        return {
            "headers": connection_params.headers,
            "base_url": connection_params.base_url,
            "query_params": connection_params.queryParams,
        }
    except ComposioClientError:
        return None


def _build_executable_from_args(  # pylint: disable=too-many-statements
    f: t.Callable,
    app: str,
) -> t.Tuple[t.Callable, t.Type[BaseModel], t.Type[BaseModel], bool]:
    """Build execute action from function arguments."""
    argspec = inspect.getfullargspec(f)
    defaults = dict(
        zip(
            reversed(argspec.annotations),
            reversed(argspec.defaults or []),
        )
    )
    docstr = getattr(f, "__doc__")
    if docstr is None:
        raise InvalidRuntimeAction(
            message=f"Runtime action `{f.__name__}` is missing docstring"
        )

    header, paramdesc, returns = _parse_docstring(
        docstr=docstr,
    )
    request_schema: t.Dict[str, t.Any] = {
        "__annotations__": {},
    }
    response_schema: t.Dict[str, t.Any] = {
        "__annotations__": {},
    }
    shell_argument = None
    auth_params = False
    request_executor = False
    if "return" not in argspec.annotations:
        raise InvalidRuntimeAction(
            f"Please add return type on runtime action `{f.__name__}`"
        )

    for arg, annot in argspec.annotations.items():
        if annot is Shell:
            shell_argument = arg
            continue

        if arg == "auth":
            auth_params = True
            continue

        if arg == "execute_request":
            request_executor = True
            continue

        if isinstance(annot, te._AnnotatedAlias):  # pylint: disable=protected-access
            annottype, description, default = _parse_annotated_type(
                argument=arg,
                annotation=annot,
            )
            default = defaults.get(arg, default)
        else:
            annottype, _ = _parse_raw_type(argument=arg, annotation=annot)
            if arg != "return" and arg not in paramdesc:
                raise InvalidRuntimeAction(
                    f"Please provide description for `{arg}` on runtime action `{f.__name__}`"
                )

            if arg == "return":
                if returns is None:
                    raise InvalidRuntimeAction(
                        f"Please provide description for `{arg}` on runtime action `{f.__name__}`"
                    )
                _, description = returns
            else:
                description = paramdesc[arg]

            default = defaults.get(arg, ...)
        if arg == "return":
            if returns is not None:
                arg, _ = returns
            else:
                arg = "result"
            response_schema[arg] = Field(default=default, description=description)
            response_schema["__annotations__"][arg] = annottype
            continue

        request_schema[arg] = Field(default=default, description=description)
        request_schema["__annotations__"][arg] = annottype

    RequestSchema = type(
        f"{inflection.camelize(f.__name__)}Request",
        (BaseModel,),
        request_schema,
    )

    ResponseSchema = type(
        f"{inflection.camelize(f.__name__)}Response",
        (BaseModel,),
        response_schema,
    )

    def execute(request: BaseModel, metadata: t.Dict) -> BaseModel:
        """Wrapper for action callable."""
        kwargs = request.model_dump()
        if shell_argument is not None:
            kwargs[shell_argument] = metadata["workspace"].shells.recent

        if auth_params > 0:
            kwargs["auth"] = (
                _get_auth_params(app=app, entity_id=metadata["entity_id"]) or {}
            )

        if request_executor:
            toolset = t.cast("ComposioToolSet", metadata["_toolset"])

            def execute_request(
                endpoint: str,
                method: str,
                body: t.Optional[t.Dict] = None,
                parameters: t.Optional[t.List[CustomAuthParameter]] = None,
            ):
                return toolset.execute_request(
                    endpoint=endpoint,
                    method=method,
                    body=body,
                    app=app,
                    parameters=parameters,
                )

            kwargs["execute_request"] = execute_request

        response = f(**kwargs)
        if isinstance(response, BaseModel):
            return response

        rname = returns[0] if returns is not None else "result"
        return ResponseSchema(**{rname: response})

    execute.__doc__ = header
    execute.__name__ = f.__name__
    return (
        execute,
        RequestSchema,
        ResponseSchema,
        shell_argument is not None,
    )


def _build_executable_from_request_class(f: t.Callable, app: str) -> t.Callable:
    def execute(request: BaseModel, metadata: t.Dict) -> BaseModel:
        """Wrapper for action callable."""
        auth_data = _get_auth_params(app=app, entity_id=metadata["entity_id"])
        if auth_data is not None:
            metadata.update(auth_data)
        return f(request, metadata)

    execute.__name__ = f.__name__
    execute.__doc__ = f.__doc__
    return execute


def _parse_schemas(
    f: t.Callable, app: str, runs_on_shell: bool
) -> t.Tuple[t.Callable, t.Type[BaseModel], t.Type[BaseModel], bool]:
    """Parse action callable schemas."""
    argspec = inspect.getfullargspec(f)
    if _is_simple_action(argspec=argspec):
        return (
            _build_executable_from_request_class(f=f, app=app),
            argspec.annotations["request"],
            argspec.annotations["return"],
            runs_on_shell,
        )
    return _build_executable_from_args(f=f, app=app)


def action(
    toolname: str,
    runs_on_shell: bool = False,
    requires: t.Optional[t.List] = None,
    tags: t.Optional[t.List[str]] = None,
) -> t.Callable[[t.Callable], t.Type[RuntimeAction]]:
    """Marks a callback as wanting to receive the current context object as first argument."""

    def wrapper(f: t.Callable) -> t.Type[RuntimeAction]:
        """Action wrapper."""
        file = inspect.getfile(f)
        f, request_schema, response_schema, _runs_on_shell = _parse_schemas(
            f=f,
            app=toolname,
            runs_on_shell=runs_on_shell,
        )
        return _wrap(
            f=f,
            toolname=toolname,
            tags=tags or [],
            file=file,
            request_schema=request_schema,
            response_schema=response_schema,
            runs_on_shell=_runs_on_shell,
            requires=requires,
        )

    return wrapper



# pylint: disable=import-outside-toplevel

from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


class ScrapeWebsiteToolRequest(BaseModel):
    url: str = Field(
        ..., description="Mandatory website url to read contents of the website from"
    )


class ScrapeWebsiteToolResponse(BaseModel):
    content: str = Field(
        ..., description="The content of the website in markdown format"
    )


class Scrape(LocalAction[ScrapeWebsiteToolRequest, ScrapeWebsiteToolResponse]):
    """
    Scrape contents of a website
    """

    _tags = ["Web"]

    def execute(
        self, request: ScrapeWebsiteToolRequest, metadata: Dict
    ) -> ScrapeWebsiteToolResponse:
        """Scrape the website and return the content in markdown format"""

        try:
            from spider import Spider
            from spider.spider_types import RequestParamsDict
        except ImportError as e:
            raise ImportError("Failed to import Spider:", e) from e

        try:
            spider = Spider()
            params = RequestParamsDict(return_format="markdown")
            response = spider.scrape_url(request.url, params)
            if response is None:
                raise ValueError(f"Error getting response for url {request.url}")
            return ScrapeWebsiteToolResponse(content=response.content)
        except (ConnectionError, TimeoutError) as e:
            raise ValueError(f"Connection or timeout error occurred: {e}") from e
        except Exception as e:
            raise ValueError(f"An unexpected error occurred: {e}") from e



from .apply_patch import ApplyPatch
from .chwdir import ChangeWorkingDirectory
from .create import CreateFile
from .edit import EditFile
from .find import FindFile
from .git_clone import GitClone
from .git_custom import GitCustom
from .git_patch import GitPatch
from .git_repo_tree import GitRepoTree
from .grep import SearchWord
from .list import ListFiles
from .open import OpenFile
from .rename import RenameFile
from .scroll import Scroll
from .write import Write



"""
Function decorators.
"""

import typing as t
import warnings

import typing_extensions as te


T = te.TypeVar("T")
P = te.ParamSpec("P")


def deprecated(
    version: str,
    replacement: str,
) -> t.Callable[[t.Callable[P, T]], t.Callable[P, T]]:
    """
    Mark a function as deprecated.

    :param version: Version where this method will be deprecated.
    :param replacement: Replacement for this method.
    """

    def wrapper(func: t.Callable[P, T]) -> t.Callable[P, T]:
        def new_func(*args: P.args, **kwargs: P.kwargs) -> T:
            warnings.warn(
                f"`{func.__name__}` is deprecated and will be removed on v{version}. "
                f"Use `{replacement}` method instead.",
                UserWarning,
            )
            return func(*args, **kwargs)

        return new_func

    return wrapper



from pathlib import Path

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


class QueryImageVectorStoreInputSchema(BaseModel):
    search_query: str = Field(..., description="Search query to retrieve the image")
    indexed_directory: str = Field(
        ...,
        description="Directory path that was indexed for image search",
    )
    max_results: int = Field(..., description="Maximum number of results to return")


class QueryImageVectorStoreOutputSchema(BaseModel):
    result: str = Field(..., description="Status of the image retrieval")
    image_paths: list[str] = Field(
        ..., description="List of retrieved image file paths"
    )


class QueryImageVectorStore(
    LocalAction[
        QueryImageVectorStoreInputSchema,
        QueryImageVectorStoreOutputSchema,
    ]
):
    """
    Query Vector Store for images
    """

    display_name = "Query Image Vector Store"
    _request_schema = QueryImageVectorStoreInputSchema
    _response_schema = QueryImageVectorStoreOutputSchema
    _tags = ["query_image_embeddings"]
    _tool_name = "embedtool"

    def execute(
        self,
        request: QueryImageVectorStoreInputSchema,
        metadata: dict = {},
    ) -> QueryImageVectorStoreOutputSchema:
        import chromadb  # pylint: disable=C0415
        from chromadb.utils import embedding_functions  # pylint: disable=C0415

        image_collection_name = Path(request.indexed_directory).name + "_images"
        index_storage_path = Path.home() / ".composio" / "image_index_storage"
        chroma_client = chromadb.PersistentClient(path=str(index_storage_path))
        chroma_collection = chroma_client.get_collection(image_collection_name)

        text_embedding_function = (
            embedding_functions.SentenceTransformerEmbeddingFunction(
                model_name="clip-ViT-B-32"
            )
        )
        query_embeddings = text_embedding_function([request.search_query])

        search_results = chroma_collection.query(
            query_embeddings=query_embeddings,
            n_results=request.max_results,
        )
        if search_results is None:
            return QueryImageVectorStoreOutputSchema(
                result="No images found", image_paths=[]
            )

        retrieved_image_paths = [
            result["file_path"] for result in search_results["metadatas"][0]
        ]

        return QueryImageVectorStoreOutputSchema(
            result="Images successfully retrieved", image_paths=retrieved_image_paths
        )



"""Tool exceptions."""

from composio.exceptions import ComposioSDKError


class ExecutionFailed(ComposioSDKError):
    """Execution failed."""

    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(message)
        self.extra = kwargs



"""Tool for executing bash commands in a persistent session."""

import typing as t

from pydantic import BaseModel, Field

from composio.tools.base.exceptions import ExecutionFailed
from composio.tools.base.local import LocalAction
from composio.tools.env.constants import EXIT_CODE, STDERR, STDOUT


class BashRequest(BaseModel):
    """Bash request abstraction."""

    command: str = Field(
        ...,
        description="Bash command to be executed.",
    )
    session_id: str = Field(
        default=None,
        description=(
            "ID of the bash session where this command will be executed. "
            "If not provided, the most recently used shell will be used to "
            "execute the command."
        ),
    )
    restart: bool = Field(
        False,
        description="If True, restart the bash session before executing the command.",
    )


class BashResponse(BaseModel):
    """Bash execution response."""

    stdout: str = Field(
        ...,
        description="Output captured from the execution of the command",
    )
    stderr: str = Field(
        ...,
        description="Errors captured during execution of the command",
    )
    exit_code: int = Field(
        ...,
        description="Exit code of the command",
    )
    session_id: t.Optional[str] = Field(
        None,
        description="ID of the bash session used for this command",
    )


class BashCommand(LocalAction[BashRequest, BashResponse]):
    """
    Run bash commands in a persistent session.

    This tool allows you to execute bash commands in a persistent session,
    maintaining state between commands. You can also restart the session
    if needed.

    Examples:
      1. Run a simple command: `ls -l`
      2. Set a variable and use it: `x=5; echo $x`
      3. Use pipes and redirections: `cat file.txt | grep "pattern" > output.txt`

    Note: Interactive commands or those requiring user input during execution
    are not supported and may cause the command to hang or timeout.
    """

    _tags = ["workspace", "bash"]

    def execute(self, request: BashRequest, metadata: t.Dict) -> BashResponse:
        """Execute a bash command."""
        try:
            if "firefox" not in request.command:
                output = self.shells.get(id=request.session_id).exec(
                    cmd=request.command
                )
                return BashResponse(
                    stdout=output[STDOUT],
                    stderr=output[STDERR],
                    exit_code=output[EXIT_CODE],
                    session_id=request.session_id,
                )

            output = self.shells.new().exec(cmd=request.command, wait=False)
            return BashResponse(
                stdout=output[STDOUT],
                stderr=output[STDERR],
                exit_code=output[EXIT_CODE],
                session_id=request.session_id,
            )
        except ExecutionFailed as e:
            return BashResponse(
                stdout="",
                stderr=str(e),
                exit_code=1,
                session_id=request.session_id,
            )



from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


# pylint: disable=import-outside-toplevel


class RagToolQueryRequest(BaseModel):
    query: str = Field(..., description="The query to search in the knowledge base")


class RagToolQueryResponse(BaseModel):
    response: str = Field(
        ..., description="The response to the query from the knowledge base"
    )


class RagToolQuery(LocalAction[RagToolQueryRequest, RagToolQueryResponse]):
    """
    Tool for querying a knowledge base.

    *Note*: This can only be performed after `add_content_to_rag_tool`
    """

    _tags = ["Knowledge Base", "rag"]

    def execute(
        self, request: RagToolQueryRequest, metadata: Dict
    ) -> RagToolQueryResponse:
        """Query the knowledge base and return the response"""
        try:
            from embedchain import App
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"Failed to import App from embedchain: {e}"
            ) from e

        try:
            embedchain_app = App()
        except Exception as e:
            raise Exception(f"Failed to initialize App: {e}") from e

        _, sources = embedchain_app.query(request.query, citations=True)
        return RagToolQueryResponse(
            response="\n\n".join([source[0] for source in sources])
        )



"""Sentry utilities."""

import atexit
import json
import os
import traceback
import types
import typing as t
from pathlib import Path

import requests
import sentry_sdk
import sentry_sdk.integrations.argv
import sentry_sdk.integrations.atexit
import sentry_sdk.integrations.dedupe
import sentry_sdk.integrations.excepthook
import sentry_sdk.integrations.fastapi
import sentry_sdk.integrations.logging
import sentry_sdk.integrations.modules
import sentry_sdk.integrations.stdlib
import sentry_sdk.integrations.threading
import sentry_sdk.types


def fetch_dsn() -> t.Optional[str]:
    request = requests.get(
        url="https://backend.composio.dev/api/v1/cli/sentry-dns",
        timeout=10,
    )
    if request.status_code != 200:
        return None
    return request.json().get("dns")


def get_sentry_config() -> t.Optional[t.Dict]:
    user_file = Path.home() / ".composio" / "user_data.json"
    if not user_file.exists():
        return None

    try:
        data = json.loads(user_file.read_text(encoding="utf-8"))
    except json.JSONDecodeError:
        return None

    if data.get("api_key") is None:
        return None

    return data.get("sentry")


def filter_sentry_errors(
    event: sentry_sdk.types.Event,
    hint: sentry_sdk.types.Hint,
) -> t.Optional[sentry_sdk.types.Event]:
    if "exc_info" not in hint:
        return None

    _, exc, trb = hint["exc_info"]
    if isinstance(exc, KeyboardInterrupt):
        return None

    trb = t.cast(types.TracebackType, trb)
    # In editable installs, we won't have composio in site-packages.
    # This ensures we don't send sentry issues during development.
    traceback_text = "".join(traceback.format_tb(trb))
    if "site-packages" + os.path.sep + "composio" not in traceback_text:
        return None

    return event


def init():
    if os.environ.get("COMPOSIO_DISABLE_SENTRY", "false").lower() in ("true", "t"):
        return

    sentry_config = get_sentry_config()
    if sentry_config is None:
        return

    if sentry_config.get("dsn") is None:
        return

    sentry_sdk.init(
        dsn=sentry_config["dsn"],
        traces_sample_rate=sentry_config.get("traces_sample_rate", 1.0),
        profiles_sample_rate=sentry_config.get("profiles_sample_rate", 1.0),
        debug=False,
        before_send=filter_sentry_errors,
        default_integrations=False,
        integrations=[
            sentry_sdk.integrations.argv.ArgvIntegration(),
            sentry_sdk.integrations.atexit.AtexitIntegration(
                callback=lambda x, y: None,
            ),  # suppress atexit message
            sentry_sdk.integrations.dedupe.DedupeIntegration(),
            sentry_sdk.integrations.excepthook.ExcepthookIntegration(),
            sentry_sdk.integrations.fastapi.FastApiIntegration(),
            sentry_sdk.integrations.logging.LoggingIntegration(),
            sentry_sdk.integrations.modules.ModulesIntegration(),
            sentry_sdk.integrations.stdlib.StdlibIntegration(),
            sentry_sdk.integrations.threading.ThreadingIntegration(),
        ],
    )


@atexit.register
def update_dsn() -> None:
    user_file = Path.home() / ".composio" / "user_data.json"
    if not user_file.exists():
        return

    try:
        data = json.loads(user_file.read_text(encoding="utf-8"))
    except json.JSONDecodeError:
        return

    if data.get("api_key") is None:
        return

    if data.get("sentry") is not None and data.get("sentry").get("dsn") is not None:
        return

    dsn = fetch_dsn()
    if dsn is None:
        return

    data["sentry"] = {"dsn": dsn}
    user_file.write_text(json.dumps(data), encoding="utf-8")



import typing as t

from pydantic import Field

from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


class ListRequest(BaseFileRequest):
    """Request to list files/directories in the workspace."""


class ListResponse(BaseFileResponse):
    """Response to list files/directories in the workspace."""

    files: t.List[t.Tuple[str, str]] = Field(
        default=[], description="List of files and directories in the workspace"
    )
    error: str = Field(default="", description="Error message if any")


class ListFiles(LocalAction[ListRequest, ListResponse]):
    """
    Lists files and directories in the current working directory opened in the
    file manager, it can result in:
    - PermissionError: If the user doesn't have permission to read the directory.
    - FileNotFoundError: If the directory does not exist.
    - OSError: If there's an issue reading the directory.
    """

    @include_cwd  # type: ignore
    def execute(self, request: ListRequest, metadata: t.Dict) -> ListResponse:
        try:
            return ListResponse(
                files=self.filemanagers.get(request.file_manager_id).ls()
            )
        except PermissionError as e:
            return ListResponse(error=f"Permission denied: {str(e)}")
        except FileNotFoundError as e:
            return ListResponse(error=f"Directory not found: {str(e)}")
        except OSError as e:
            return ListResponse(error=f"OS error occurred: {str(e)}")



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateTaskAttachmentRequest(BaseModel):
    """Request schema for `CreateTaskAttachment`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )


class CreateTaskAttachmentResponse(BaseModel):
    """Response schema for `CreateTaskAttachment`"""

    data: t.Dict[str, t.Any]


class CreateTaskAttachment(OpenAPIAction):
    """
    Upload a file to a task as an attachment.    ***Note:** This request uses
    multipart/form-data as the content type.*
    """

    _tags = ["Attachments"]
    _display_name = "create_task_attachment"
    _request_schema = CreateTaskAttachmentRequest
    _response_schema = CreateTaskAttachmentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/attachment"
    method = "post"
    operation_id = "Attachments_uploadFileToTaskAsAttachment"
    action_identifier = "/task/{task_id}/attachment_post"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {}

    aliases = {}


class attachments_upload_file_to_task_as_attachment(OpenAPIAction):
    """
    Upload a file to a task as an attachment.    ***Note:** This request uses
    multipart/form-data as the content type.*<<DEPRECATED use create_task_attachment>>
    """

    _tags = ["Attachments"]
    _display_name = "attachments_upload_file_to_task_as_attachment"
    _request_schema = CreateTaskAttachmentRequest
    _response_schema = CreateTaskAttachmentResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/attachment"
    method = "post"
    operation_id = "Attachments_uploadFileToTaskAsAttachment"
    action_identifier = "/task/{task_id}/attachment_post"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {}

    aliases = {}
